package ch.epfl.lamp.minischeme;

import java.io._;

abstract class Constant;
case class IntegerConstant(value: Int) extends Constant {
  override def toString(): String = value.toString();
}
case class LabelConstant(value: Label) extends Constant {
  override def toString(): String = value.toString();
}

abstract class Instruction(val opcode: Opcode) {
  def printTo(stream: PrintStream): Unit = {
    stream.println("\t" + opcode + " " + argsAsString);
  }

  protected def argsAsString: String;
}

class InstructionRRR(_op: OpcodeRRR, _r1: Register, _r2: Register, _r3: Register) extends Instruction(_op) {
  var r1: Register = _r1;
  var r2: Register = _r2;
  var r3: Register = _r3;
  override def argsAsString: String =
    "" + r1 + " " + r2 + " " + r3;
}

class InstructionRR(_op: ALOC.type, _r1: Register, _r2: Register) extends Instruction(_op) {
  var r1: Register = _r1;
  var r2: Register = _r2;
  override def argsAsString: String =
    "" + r1 + " " + r2;
}

class InstructionRRC(_op: OpcodeRRC, _r1: Register, _r2: Register, _constant: Constant) extends Instruction(_op) {
  var r1: Register = _r1;
  var r2: Register = _r2;
  var constant: Constant = _constant;
  override def argsAsString: String =
    "" + r1 + " " + r2 + " " + constant;
}

class InstructionRC(_op: LINT.type, _r: Register, _constant: Constant) extends Instruction(_op) {
  var r: Register = _r;
  var constant: Constant = _constant;
  override def argsAsString: String =
    "" + r + " " + constant;
}

class InstructionR(_op: OpcodeR, _r: Register) extends Instruction(_op) {
  var r: Register = _r;
  override def argsAsString: String = r.toString();
}

class InstructionHalt(_op: HALT.type) extends Instruction(_op) {
  override def argsAsString: String = "";
}
