package ch.epfl.lamp.minischeme;

import java.io._;

/**
 * Parser for minischeme.
 *
 * @author Michel Schinz
 */
class Parser(scanner: Scanner) {
  var nextToken: Token = scanner.nextToken;
  def skipToken: Unit =
    nextToken = scanner.nextToken;

  // program ::= { expr }
  def program: List[Tree] = listOfExpr;

  // expr ::= list | symbol | number
  private def expr: Tree = setPos { nextToken match {
    case LeftParen() => list
    case SymbolToken(n) => skipToken; SIdent(n)
    case NumberToken(v) => skipToken; SNum(v)
    case StringToken(s) => skipToken; stringList(s);
    case _ => Report.error(nextToken.line,
                           nextToken.column,
                           "unexpected token " + nextToken)
  }}

  // list ::= '(' { expr } ')'
  private def list: Tree = setPos {
    skipToken;
    val res = listOfExpr;
    if (nextToken != RightParen())
      Report.error(nextToken.line,
                   nextToken.column,
                   "expected ')', found " + nextToken);
    skipToken;
    SList(res)
  }

  private def listOfExpr: List[Tree] = nextToken match {
    case LeftParen() | NumberToken(_) | SymbolToken(_) | StringToken(_) =>
      val head = expr;
      head :: listOfExpr
    case _ =>
      List()
  }

  private def stringList(s: String): Tree = {
    var res: Tree = SNum(0);
    var i = s.length() - 1;
    while (i >= 0) {
      res = SList(SIdent("cons") :: SNum(s.charAt(i)) :: res :: Nil);
      i = i - 1;
    }
    res
  }

  private def setPos(body: =>Tree): Tree = {
    val tok = nextToken;
    val tree = body;
    tree.line = tok.line;
    tree.column = tok.column;
    tree
  }
}
