package ch.epfl.lamp.minischeme;

import java.io._;

/**
 * Scanner for minischeme.
 *
 * @author Michel Schinz
 */
class Scanner(inputStreams: List[InputStream]) {
  private var nextCharAsInt : Int = _;
  private var line, column : Int = 1;
  def nextChar: Char = nextCharAsInt.asInstanceOf[Char];

  private var inputStreamReaders: List[InputStreamReader] =
    inputStreams map { s => new InputStreamReader(s) };
  private def skipChar: Unit = {
    nextCharAsInt = inputStreamReaders.head.read();
    if (nextCharAsInt == -1) {
      line = 1;
      column = 1;
      inputStreamReaders = inputStreamReaders.tail;
      if (!inputStreamReaders.isEmpty) skipChar;
    } else if (nextCharAsInt.asInstanceOf[Char] == '\n') {
      line = line + 1;
      column = 1;
    } else
      column = column + 1;
  }
  skipChar;

  def nextToken: Token = {
    val l = line; val c = column;
    val token = readNextToken;
    token.line = l; token.column = c;
    token
  }

  private def readNextToken : Token = {
    if (nextCharAsInt == -1)
      EOF()
    else {
      if (Character.isWhitespace(nextChar)) {
        while (Character.isWhitespace(nextChar)) skipChar;
        readNextToken
      } else if (nextChar == ';') {
        while (nextChar != '\n') skipChar;
        readNextToken
      } else if (isNumberCharacter(nextChar)) {
        readNumber
      } else if (isSymbolStartCharacter(nextChar)) {
        readSymbol
      } else if (nextChar == '"') {
        readString
      } else nextChar match {
          case '(' => skipChar; LeftParen()
          case ')' => skipChar; RightParen()
          case _ => Report.error(line, column, "Unknown character: '" + nextChar + "'")
      }
    }
  }

  private val numberCharacters = "0123456789";
  def isNumberCharacter(c: Char): Boolean =
    (numberCharacters indexOf c) != -1;
  private val symbolStartCharacters =
    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ|!$%&*+-./:<=>?@^_~";
  def isSymbolStartCharacter(c: Char): Boolean =
    (symbolStartCharacters indexOf c) != -1;
  def isSymbolComponentCharacter(c: Char): Boolean =
    isSymbolStartCharacter(c) || isNumberCharacter(c);

  def readSymbol: SymbolToken =
    SymbolToken(readStringWhile(isSymbolComponentCharacter));
  def readNumber: NumberToken =
    NumberToken(Integer.parseInt(readStringWhile(isNumberCharacter)));
  def readString: StringToken = {
    skipChar;
    val str = readStringWhile(c => c != '"');
    skipChar;
    StringToken(str)
  }

  def readStringWhile(predicate: Char=>Boolean): String = {
    val buffer = new StringBuffer();
    do {
      buffer append nextChar;
      skipChar
    } while (nextCharAsInt != -1 && predicate(nextChar));
    buffer.toString()
  }
}
