package ch.epfl.lamp.minischeme;

import java.io.PrintWriter;
import scala.text.Document;

/**
 * Format the tree on 80 columns.
 *
 * @author Michel Schinz
 */
object TreePrinter extends Function1[Tree,Unit] {
  def apply(tree: Tree): Unit = {
    val doc = toDocument(tree);
    val writer = new PrintWriter(System.out);
    doc.format(80, writer);
    writer.println();
    writer.flush();
  }

  import Document._;

  private def toDocument(tree: Tree): Document = tree match {
    case SList(SIdent("define") :: SIdent(name) :: value :: Nil) =>
      "(define " :: nest(2, name :/: group(toDocument(value))) :: ")" :: empty
    case SList(SIdent("if") :: cond :: thenp :: elsep :: Nil) =>
      "(if " :: nest(2, toDocument(cond)
                     :/: toDocument(thenp)
                     :/: toDocument(elsep)) :: ")" :: empty
    case SList(SIdent("let") :: SList(bindings) :: body) =>
      "(let " :: nest(2, "(" :: group(toDocument(bindings)) :: ")"
                      :/: group(toDocument(body)) :: ")"
                      :: empty)
    case SList(SIdent("lambda") :: params :: body) =>
      "(lambda " :: nest(2, toDocument(params)
                         :/: group(toDocument(body))) :: ")" :: empty
    case SList(elements) =>
      nest(1,"(" :: group(toDocument(elements)) :: ")" :: empty)
    case SIdent(name) =>
      Document.text(name)
    case SNum(value) =>
      Document.text(String.valueOf(value))
  }

  private def toDocument(trees: List[Tree]): Document = trees match {
    case first :: second :: rest =>
      toDocument(first) :/: rest.foldLeft(toDocument(second)) { (d,e) =>
        d :/: toDocument(e)
      }
    case single :: Nil => toDocument(single)
    case _ => empty
  }
}
